/*
 * Copyright (C) 2002 Bjrn Englund
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "userscsi.h"

#include "userscsi_str.h"




#define REQ_SENSE_LEN 18
uint8_t request_sense_data[REQ_SENSE_LEN];
int request_sense_data_len = REQ_SENSE_LEN;
uint8_t status;

typedef struct {
  uint8_t b[4];
} StructureListEntry_t;

typedef struct {
  uint8_t b[2052];
} PhysicalFormatInformation_t;

typedef struct {
  uint8_t b[8];
} CopyrightInformation_t;

typedef struct {
  uint8_t b[2052];
} DiscManufacturingInformation_t;

typedef struct {
  uint8_t b[8];
} CapacityData_t;



int ReportKey_RPCState(int fd,
		       uint8_t *data, uint16_t data_len)
{
  userscsi_cmd_t cmd;
  uint8_t cdb[12];
  int ret;

  memset(cdb, 0, sizeof(cdb));
  memset(request_sense_data, 0, request_sense_data_len);

  cdb[0] = REPORT_KEY;

  cdb[8] = data_len >> 8;
  cdb[9] = data_len & 0xff;

  cdb[10] = REPORT_KEY_FORMAT_RPC_STATE & 0x3f;
  
  cmd.flags = USERSCSI_CMD_READ | USERSCSI_CMD_SILENT | USERSCSI_CMD_RQENABLE;
  cmd.cdb = cdb;
  cmd.cdb_len = 12;
  cmd.data = data;
  cmd.data_len = data_len;
  cmd.rs = request_sense_data;
  cmd.rs_len = request_sense_data_len;
  cmd.timeout = 5;

  
  ret = userscsi_cmd(fd, &cmd);
  
  status = cmd.status;
  
  return ret;
}



void print_RPCState(FILE *fp, uint8_t *data, int data_len)
{

  fprintf(fp, "Report Key Parameter List - RPC State\n");
  fprintf(fp, "Type Code: %01x\n", data[4]>>6);
  fprintf(fp, "Number of Vendor Resets Available: %01x\n", data[4]>>3 & 0x07);
  fprintf(fp, "Number of User Controlled Changes Available: %01x\n", data[4]&0x07);

  fprintf(fp, "Region Mask: %02x ", data[5]);
  fprintf(fp, "RPC Scheme: %02x\n", data[6]);

}

int ReadDVDStructure_StructureList(int fd, int layer_nr,
				  uint8_t *data, uint16_t data_len)
{
  userscsi_cmd_t cmd;
  uint8_t cdb[12];
  int ret;

  memset(cdb, 0, sizeof(cdb));
  memset(request_sense_data, 0, request_sense_data_len);

  cdb[0] = READ_DVD_STRUCTURE;
  cdb[6] = layer_nr; // layer 0-1
  cdb[7] = READ_DVD_STRUCTURE_FORMAT_STRUCTURE_LIST;
  cdb[8] = data_len >> 8;
  cdb[9] = data_len & 0xff;

  cmd.flags = USERSCSI_CMD_READ | USERSCSI_CMD_SILENT | USERSCSI_CMD_RQENABLE;
  cmd.cdb = cdb;
  cmd.cdb_len = 12;
  cmd.data = data;
  cmd.data_len = data_len;
  cmd.rs = request_sense_data;
  cmd.rs_len = request_sense_data_len;
  cmd.timeout = 5;

  
  ret = userscsi_cmd(fd, &cmd);
  
  status = cmd.status;
  
  return ret;
}

int ReadCapacity(int fd, uint8_t *data, uint16_t data_len)
{
  userscsi_cmd_t cmd;
  uint8_t cdb[12];
  int ret;

  memset(cdb, 0, sizeof(cdb));
  memset(request_sense_data, 0, request_sense_data_len);

  cdb[0] = READ_CAPACITY;

  cmd.flags = USERSCSI_CMD_READ | USERSCSI_CMD_SILENT | USERSCSI_CMD_RQENABLE;
  cmd.cdb = cdb;
  cmd.cdb_len = 10;
  cmd.data = data;
  cmd.data_len = data_len;
  cmd.rs = request_sense_data;
  cmd.rs_len = request_sense_data_len;
  cmd.timeout = 5;

  
  ret = userscsi_cmd(fd, &cmd);
  
  status = cmd.status;
  
  return ret;
}


void print_CapacityData(FILE *fp,
			CapacityData_t *capacity)
{
  fprintf(fp, "Capacity Data\n");
  
  fprintf(fp, "Logical Block Address: %08x\n",
	  capacity->b[0]<<24 |
	  capacity->b[1]<<16 |
	  capacity->b[2]<<8  |
	  capacity->b[3]);

  fprintf(fp, "Block Length: %08x\n",
	  capacity->b[4]<<24 |
	  capacity->b[5]<<16 |
	  capacity->b[6]<<8  |
	  capacity->b[7]);

}


int Read_10(int fd, uint32_t lba, uint16_t nr_lba,  int fua,
	    uint8_t *data, int data_len)
{
  userscsi_cmd_t cmd;
  uint8_t cdb[12];
  int ret;
  
  memset(cdb, 0, sizeof(cdb));
  memset(request_sense_data, 0, request_sense_data_len);

  if(nr_lba*2048 > data_len) {
    nr_lba = data_len/2048;
    fprintf(stdout, "ERR: Read_10, not enough allocated space for blocks\n");
    fprintf(stdout, "ERR: lowering to %d blocks\n", nr_lba);
  }
  
  cdb[0] = READ_10;
  cdb[1] = (fua & 0x1)<<3;
  cdb[2] = (lba >> 24) & 0xff;
  cdb[3] = (lba >> 16) & 0xff;
  cdb[4] = (lba >> 8) & 0xff;
  cdb[5] = lba & 0xff;

  cdb[7] = (nr_lba >> 8) & 0xff;
  cdb[8] = nr_lba & 0xff;

  cmd.flags = USERSCSI_CMD_READ | USERSCSI_CMD_SILENT | USERSCSI_CMD_RQENABLE;
  cmd.cdb = cdb;
  cmd.cdb_len = 10;
  cmd.data = (uint8_t *)data;
  cmd.data_len = data_len;
  cmd.rs = request_sense_data;
  cmd.rs_len = request_sense_data_len;
  cmd.timeout = 5;

  
  ret = userscsi_cmd(fd, &cmd);

  status = cmd.status;
  
  return ret;
}


int Read_12(int fd, uint32_t lba, uint32_t nr_lba,  int fua,
	    uint8_t *data, int data_len)
{
  userscsi_cmd_t cmd;
  uint8_t cdb[12];
  int ret;
  
  memset(cdb, 0, sizeof(cdb));
  memset(request_sense_data, 0, request_sense_data_len);

  if(nr_lba*2048 > data_len) {
    nr_lba = data_len/2048;
    fprintf(stdout, "ERR: Read_12, not enough allocated space for blocks\n");
    fprintf(stdout, "ERR: lowering to %d blocks\n", nr_lba);
  }
  
  cdb[0] = READ_12;
  cdb[1] = (fua & 0x1)<<3;
  cdb[2] = (lba >> 24) & 0xff;
  cdb[3] = (lba >> 16) & 0xff;
  cdb[4] = (lba >> 8) & 0xff;
  cdb[5] = lba & 0xff;

  cdb[6] = (nr_lba >> 24) & 0xff;
  cdb[7] = (nr_lba >> 16) & 0xff;
  cdb[8] = (nr_lba >> 8) & 0xff;
  cdb[9] = nr_lba & 0xff;

  cmd.flags = USERSCSI_CMD_READ | USERSCSI_CMD_SILENT | USERSCSI_CMD_RQENABLE;
  cmd.cdb = cdb;
  cmd.cdb_len = 12;
  cmd.data = (uint8_t *)data;
  cmd.data_len = data_len;
  cmd.rs = request_sense_data;
  cmd.rs_len = request_sense_data_len;
  cmd.timeout = 5;

  
  ret = userscsi_cmd(fd, &cmd);

  status = cmd.status;
  
  return ret;
}


int ReadDVDStructure_Physical(int fd, int layer_nr,
			      PhysicalFormatInformation_t *data)
{
  userscsi_cmd_t cmd;
  uint8_t cdb[12];
  int ret;
  
  memset(cdb, 0, sizeof(cdb));
  memset(request_sense_data, 0, request_sense_data_len);

  cdb[0] = READ_DVD_STRUCTURE;
  cdb[6] = layer_nr; // layer 0-1
  cdb[7] = READ_DVD_STRUCTURE_FORMAT_PHYSICAL;
  cdb[8] = sizeof(*data) >> 8;
  cdb[9] = sizeof(*data) & 0xff;

  cmd.flags = USERSCSI_CMD_READ | USERSCSI_CMD_SILENT | USERSCSI_CMD_RQENABLE;
  cmd.cdb = cdb;
  cmd.cdb_len = 12;
  cmd.data = (uint8_t *)data;
  cmd.data_len = sizeof(*data);
  cmd.rs = request_sense_data;
  cmd.rs_len = request_sense_data_len;
  cmd.timeout = 5;

  
  ret = userscsi_cmd(fd, &cmd);

  status = cmd.status;
  
  return ret;
}


int ReadDVDStructure_Copyright(int fd, int layer_nr,
			       CopyrightInformation_t *data)
{
  userscsi_cmd_t cmd;
  uint8_t cdb[12];
  int ret;
  
  memset(cdb, 0, sizeof(cdb));
  memset(request_sense_data, 0, request_sense_data_len);

  cdb[0] = READ_DVD_STRUCTURE;
  cdb[6] = layer_nr; // layer 0-1
  cdb[7] = READ_DVD_STRUCTURE_FORMAT_COPYRIGHT;
  cdb[8] = sizeof(*data) >> 8;
  cdb[9] = sizeof(*data) & 0xff;

  cmd.flags = USERSCSI_CMD_READ | USERSCSI_CMD_SILENT | USERSCSI_CMD_RQENABLE;
  cmd.cdb = cdb;
  cmd.cdb_len = 12;
  cmd.data = (uint8_t *)data;
  cmd.data_len = sizeof(*data);
  cmd.rs = request_sense_data;
  cmd.rs_len = request_sense_data_len;
  cmd.timeout = 5;

  
  ret = userscsi_cmd(fd, &cmd);

  status = cmd.status;

  return ret;
}


int ReadDVDStructure_Manufacturing(int fd, int layer_nr,
				   DiscManufacturingInformation_t *data)
{
  userscsi_cmd_t cmd;
  uint8_t cdb[12];
  int ret;
  
  memset(cdb, 0, sizeof(cdb));
  memset(request_sense_data, 0, request_sense_data_len);

  cdb[0] = READ_DVD_STRUCTURE;
  cdb[6] = layer_nr; // layer 0-1
  cdb[7] = READ_DVD_STRUCTURE_FORMAT_MANUFACTURING;
  cdb[8] = sizeof(*data) >> 8;
  cdb[9] = sizeof(*data) & 0xff;

  cmd.flags = USERSCSI_CMD_READ | USERSCSI_CMD_SILENT | USERSCSI_CMD_RQENABLE;
  cmd.cdb = cdb;
  cmd.cdb_len = 12;
  cmd.data = (uint8_t *)data;
  cmd.data_len = sizeof(*data);
  cmd.rs = request_sense_data;
  cmd.rs_len = request_sense_data_len;
  cmd.timeout = 5;

  
  ret = userscsi_cmd(fd, &cmd);

  status = cmd.status;

  return ret;
}


int GetConfiguration(int fd, int rt, int sfn,
		     uint8_t *data, int data_len)
{
  userscsi_cmd_t cmd;
  uint8_t cdb[10];
  int ret;
  
  memset(cdb, 0, sizeof(cdb));
  memset(request_sense_data, 0, request_sense_data_len);

  cdb[0] = GET_CONFIGURATION;
  cdb[1] = rt & 0x3;
  cdb[2] = (sfn >> 8) & 0xff;
  cdb[3] = sfn & 0xff;

  cdb[7] = data_len >> 8;
  cdb[8] = data_len & 0xff;

  cmd.flags = USERSCSI_CMD_READ | USERSCSI_CMD_SILENT | USERSCSI_CMD_RQENABLE;
  cmd.cdb = cdb;
  cmd.cdb_len = sizeof(cdb);
  cmd.data = data;
  cmd.data_len = data_len;
  cmd.rs = request_sense_data;
  cmd.rs_len = request_sense_data_len;
  cmd.timeout = 5;

  
  ret = userscsi_cmd(fd, &cmd);

  status = cmd.status;
  
  return ret;
}

const char *get_Profile_str(uint16_t profile)
{
  char *ret_str;

  switch(profile) {
  case  0x0001:
    ret_str = "Non-removable disk";
    break;
  case 0x0002:
    ret_str = "Removable disk";
    break;
  case 0x0003:
    ret_str = "MO Erasable";
    break;
  case 0x0004:
    ret_str = "MO Write Once";
    break;
  case 0x0005:
    ret_str = "AS-MO";
    break;
  case 0x0008:
    ret_str = "CD-ROM";
    break;
  case 0x0009:
    ret_str = "CD-R";
    break;
  case 0x000A:
    ret_str = "CD-RW";
    break;
  case 0x0010:
    ret_str = "DVD-ROM";
    break;
  case 0x0011:
    ret_str = "DVD-R Sequential Recording";
    break;
  case 0x0012:
    ret_str = "DVD-RAM";
    break;
  case 0x0013:
    ret_str = "DVD-RW Restricted Overwrite";
    break;
  case 0x0014:
    ret_str = "DVD-RW Sequential recording";
    break;
  case 0x001A:
    ret_str = "DVD+RW";
    break;
  case 0x0020:
    ret_str = "DDCD-ROM";
    break;
  case 0x0021:
    ret_str = "DDCD-R";
    break;
  case 0x0022:
    ret_str = "DDCD-RW";
    break;
  case 0xFFFF:
    ret_str = "Logical Unit Not Conforming to a Standard Profile";
    break;
  default:
    ret_str = "Reserved";
    break;
  }

  return ret_str;
}


const char *get_Feature_str(uint16_t feature)
{
  char *ret_str;


  if((feature >= 0xFF00) && (feature <= 0xFFFF)) {
    return "Vendor Specific";
  }

  switch(feature) {
  case  0x0000:
    ret_str = "Profile List";
    break;
  case  0x0001:
    ret_str = "Core";
    break;
  case 0x0002:
    ret_str = "Morphing";
    break;
  case 0x0003:
    ret_str = "Removable Medium";
    break;
  case 0x0004:
    ret_str = "Write Protect";
    break;
  case 0x0010:
    ret_str = "Random Readable";
    break;
  case 0x001D:
    ret_str = "Multi-Read";
    break;
  case 0x001E:
    ret_str = "CD Read";
    break;
  case 0x001F:
    ret_str = "DVD Read";
    break;
  case 0x0020:
    ret_str = "Random Writable";
    break;
  case 0x0021:
    ret_str = "Incremental Streaming Writable";
    break;
  case 0x0022:
    ret_str = "Sector erasable";
    break;
  case 0x0023:
    ret_str = "Formattable";
    break;
  case 0x0024:
    ret_str = "Defect Management";
    break;
  case 0x0025:
    ret_str = "Write Once";
    break;
  case 0x0026:
    ret_str = "Restricted Overwrite";
    break;
  case 0x0027:
    ret_str = "CD-RW CAV Write";
    break;
  case 0x0028:
    ret_str = "MRW";
    break;
  case 0x002A:
    ret_str = "DVD+RW";
    break;
  case 0x002C:
    ret_str = "Rigid Restricted Overwrite";
    break;
  case 0x002D:
    ret_str = "CD Track at Once";
    break;
  case 0x002E:
    ret_str = "CD Mastering";
    break;
  case 0x002F:
    ret_str = "DVD-R/-RW Write";
    break;
  case 0x0030:
    ret_str = "DDCD Read";
    break;
  case 0x0031:
    ret_str = "DDCD-R Write";
    break;
  case 0x0032:
    ret_str = "DDCD-RW Write";
    break;
  case 0x0100:
    ret_str = "Power Management";
    break;
  case 0x0101:
    ret_str = "S.M.A.R.T.";
    break;
  case 0x0102:
    ret_str = "Embedded Changer";
    break;
  case 0x0103:
    ret_str = "CD Audio analog play";
    break;
  case 0x0104:
    ret_str = "Microcode Upgrade";
    break;
  case 0x0105:
    ret_str = "Time-out";
    break;
  case 0x0106:
    ret_str = "DVD-CSS";
    break;
  case 0x0107:
    ret_str = "Real Time Streaming";
    break;
  case 0x0108:
    ret_str = "Logical Unit serial number";
    break;
  case 0x010A:
    ret_str = "Disc Control Blocks";
    break;
  case 0x010B:
    ret_str = "DVD CPRM";
    break;
  default:
    ret_str = "Reserved";
    break;
  }

  return ret_str;
}

const char *get_PhysicalInterfaceStandard_str(uint32_t interface)
{
  const char *ret_str;
  
  switch(interface) {
  case 0:
    ret_str = "Unspecified";
    break;
  case 1:
    ret_str = "SCSI Family";
    break;
  case 2:
    ret_str = "ATAPI";
    break;
  case 3:
    ret_str = "IEEE 1394-1995";
    break;
  case 4:
    ret_str = "IEEE 1394A";
    break;
  case 5:
    ret_str = "Fibre Channel";
    break;
  case 0xFFFF:
    ret_str = "Vendor Unique";
    break;
  default:
    if(interface >= 0x10000 && interface < 0x20000) {
      ret_str = "Defined by NCITS";
    } else if(interface >= 0x20000 && interface < 0x30000) {
      ret_str = "Defined by SFF";
    } if(interface >= 0x30000 && interface < 0x40000) {
      ret_str = "Defined by IEEE";
    } else {
      ret_str = "Reserved";
    }
    break;
  }

  return ret_str;
}

const char *get_LoadingMechanismType_str(uint8_t loading_mechansim)
{
  const char *ret_str;
  
  switch(loading_mechansim) {
  case 0x0:
    ret_str = "Caddy/Slot";
    break;
  case 0x1:
    ret_str = "Tray";
    break;
  case 0x2:
    ret_str = "Pop-up";
    break;
  case 0x4:
    ret_str = "Embedded changer with individually changeable discs";
    break;
  case 0x5:
    ret_str = "Embedded changer using a magazine mechanism";
    break;
  default:
    ret_str = "Reserved";
    break;    
  }
  return ret_str;
}

int print_FeatureDescriptor(FILE *fp, uint8_t *data, int data_len)
{
  int n = 0;
  int additional_len;
  uint16_t feature_code;
  int i;
  
  if(data_len < 4)
    return data_len;
  
  feature_code = data[n]<<8 | data[n+1];
  fprintf(fp, "Feature Code: (%#04x) %s\n",
	  feature_code,
	  get_Feature_str(feature_code));
  fprintf(fp, "Version: %#01x, Persistent: %01x, Current: %01x\n",
	  (data[n+2]>>2) & 0x0f,
	  (data[n+2]>>1) & 0x01,
	  data[n+2] & 0x01);
  additional_len = data[n+3];
  n+=4;
  switch(feature_code) {
  case 0x0000:
    for(i = 0; i < additional_len/4; i++) {
      if(data_len >= n+4) {
	fprintf(fp, "Profile Number: (%#04x) %s, current: %01x\n",
		data[n]<<8 | data[n+1],
		get_Profile_str(data[n]<<8 | data[n+1]),
		data[n+2] & 0x1);
	n+=4;
      } else {
	break;
      }
    }
    break;
  case 0x0001:
    if(data_len >= n+4) {
      uint32_t pif;
      pif = data[n]<<24 | data[n+1]<<16 | data[n+2]<<8 | data[n+3];
      fprintf(fp, "Physical Interface Standard: (%#08x) %s\n",
	      pif,
	      get_PhysicalInterfaceStandard_str(pif));
    }
    n+=additional_len;
    break;
  case 0x0002:
    if(data_len >= n+1) {
      fprintf(fp, "ASYNC: %01x\n", data[n] & 0x01);
    }
    n+= additional_len;
    break;
  case 0x0003:
    if(data_len >= n+1) {
      fprintf(fp, "Loading Mechanism Type: (%01x) %s\n",
	      (data[n]>>5) & 0x7,
	      get_LoadingMechanismType_str((data[n]>>5) & 0x7));
      fprintf(fp, "Eject: %01x, Pvnt Jmpr: %01x, Lock: %01x\n",
	      (data[n]>>3) & 0x1,
	      (data[n]>>2) & 0x1,
	      data[n] & 0x1);
    }
    n+= additional_len;
    break;
  default:
    n+=additional_len;
    break;
  }
  return n;
}


void print_ConfigurationData(FILE *fp, uint8_t *data, int data_size)
{
  int data_len;
  int n;
  
  fprintf(fp, "Configuration Data\n");
  if(data_size < 4)
    return;
  
  data_len = data[0]<<24 | data[1]<<16 | data[2]<<8 | data[3];

  fprintf(fp, "Data Length: %#08x\n",
	  data_len);

  if(data_len+4 > data_size) {
    fprintf(fp, "Data truncated (size %#08x)\n", data_size);
    data_len = data_size;
  } else {
    data_len+=4;
  }

  n = 6;
  if(data_len < 8)
    return;
  
  fprintf(fp, "Current Profile: (%#04x) %s\n",
	  data[n]<<8 | data[n+1],
	  get_Profile_str(data[n]<<8 | data[n+1]));
  
  n+=2;
  
  do {
    n+= print_FeatureDescriptor(fp, &data[n], data_len-n);
  } while(data_len > n);
}

int Inquiry(int fd, uint8_t *data, uint8_t data_len)
{
  userscsi_cmd_t cmd;
  uint8_t cdb[6];
  int ret;
  
  memset(cdb, 0, sizeof(cdb));
  memset(request_sense_data, 0, request_sense_data_len);

  cdb[0] = INQUIRY;
  cdb[4] = data_len;

  cmd.flags = USERSCSI_CMD_READ | USERSCSI_CMD_SILENT | USERSCSI_CMD_RQENABLE;
  cmd.cdb = cdb;
  cmd.cdb_len = sizeof(cdb);
  cmd.data = data;
  cmd.data_len = data_len;
  cmd.rs = request_sense_data;
  cmd.rs_len = request_sense_data_len;
  cmd.timeout = 5;

  
  ret = userscsi_cmd(fd, &cmd);

  status = cmd.status;

  return ret;
}


int Inquiry_CmdDt(int fd, uint8_t *data, uint8_t data_len, uint8_t op_code)
{
  userscsi_cmd_t cmd;
  uint8_t cdb[6];
  int ret;
  
  memset(cdb, 0, sizeof(cdb));
  memset(request_sense_data, 0, request_sense_data_len);

  cdb[0] = INQUIRY;
  cdb[1] = 0x02;
  cdb[2] = op_code;
  cdb[4] = data_len;

  cmd.flags = USERSCSI_CMD_READ | USERSCSI_CMD_SILENT | USERSCSI_CMD_RQENABLE;
  cmd.cdb = cdb;
  cmd.cdb_len = sizeof(cdb);
  cmd.data = data;
  cmd.data_len = data_len;
  cmd.rs = request_sense_data;
  cmd.rs_len = request_sense_data_len;
  cmd.timeout = 5;

  
  ret = userscsi_cmd(fd, &cmd);

  status = cmd.status;

  return ret;
}


void print_CommandSupportData(FILE *fp, uint8_t *data, int data_len)
{

  fprintf(fp, "Command Support Data\n");
  fprintf(fp, "Peripheral Qualifier: %01x\n", data[0]>>5);
  fprintf(fp, "Peripheral Device Type: %02x", data[0] & 0x1f);
  switch(data[0] & 0x1f) {
  case 0x00:
    fprintf(fp, " Direct-access (magnetic disk)\n");
    break;
  case 0x01:
    fprintf(fp, " Sequential-access (magnetic tape)\n");
    break;
  case 0x02:
    fprintf(fp, " Printer\n");
    break;
  case 0x03:
    fprintf(fp, " Processor\n");
    break;
  case 0x04:
    fprintf(fp, " Write-once (some optical disks)\n");
    break;
  case 0x05:
    fprintf(fp, " C/DVD (Multimedia)\n");
    break;
  case 0x06:
    fprintf(fp, " Scanner (obsolete)\n");
    break;
  case 0x07:
    fprintf(fp, " Optical memory (some optical disks)\n");
    break;
  case 0x08:
    fprintf(fp, " Medium changer (jukeboxes)\n");
    break;
  case 0x09:
    fprintf(fp, " Communications (obsolete)\n");
    break;
  case 0x0A:
  case 0x0B:
    fprintf(fp, " (Graphics arts pre-press)\n");
    break;
  case 0x0C:
    fprintf(fp, " Storage array controller (RAID)\n");
    break;
  case 0x0D:
    fprintf(fp, " Enclosure services\n");
    break;
  case 0x0E:
    fprintf(fp, " Simplified direct-access (magnetic disk)\n");
    break;
  case 0x0F:
    fprintf(fp, " Optical card reader/writer\n");
    break;
  case 0x10:
    fprintf(fp, " (Reserved by Bridging Expanders)\n");
    break;
  case 0x11:
    fprintf(fp, " Object-based storage\n");
    break;
  case 0x12:
    fprintf(fp, " Automation/Drive Interface\n");
    break;
  case 0x1E:
    fprintf(fp, " Well known logical unit\n");
    break;
  case 0x1F:
    fprintf(fp, " Unknown or no device type\n");
    break;
  default:
    fprintf(fp, " Reserved\n");
    break;
    
  }

  fprintf(fp, "Support:");
  switch(data[1]&0x07) {
  case 0x0:
    fprintf(fp, "Data not currently available\n");
    break;
  case 0x1:
    fprintf(fp, "Operation code not supported\n");
    break;
  case 0x2:
    fprintf(fp, "(0x2) Reserved");
    break;
  case 0x3:
    fprintf(fp, "Operation code supported in conformance with a SCSI standard\n");
    break;
  case 0x4:
    fprintf(fp, "(0x4) Vendor specific\n");
    break;
  case 0x5:
    fprintf(fp, "Operation code supported in a vendor specific manner");
    break;
  case 0x6:
    fprintf(fp, "(0x6) Vendor specific\n");
    break;
  case 0x7:
    fprintf(fp, "(0x7) Reserved\n");
    break;
  }
  
  fprintf(fp, "Version: %02x\n", data[2]);
  
  fprintf(fp, "CDB Size: %02x\n", data[5]);

}


void print_InquiryData(FILE *fp, uint8_t *data, int data_len)
{

  fprintf(fp, "Inquiry data\n");
  fprintf(fp, "Peripheral Qualifier: %01x\n", data[0]>>5);
  fprintf(fp, "Peripheral Device Type: %02x", data[0] & 0x1f);
  switch(data[0] & 0x1f) {
  case 0x00:
    fprintf(fp, " Direct-access (magnetic disk)\n");
    break;
  case 0x01:
    fprintf(fp, " Sequential-access (magnetic tape)\n");
    break;
  case 0x02:
    fprintf(fp, " Printer\n");
    break;
  case 0x03:
    fprintf(fp, " Processor\n");
    break;
  case 0x04:
    fprintf(fp, " Write-once (some optical disks)\n");
    break;
  case 0x05:
    fprintf(fp, " C/DVD (Multimedia)\n");
    break;
  case 0x06:
    fprintf(fp, " Scanner (obsolete)\n");
    break;
  case 0x07:
    fprintf(fp, " Optical memory (some optical disks)\n");
    break;
  case 0x08:
    fprintf(fp, " Medium changer (jukeboxes)\n");
    break;
  case 0x09:
    fprintf(fp, " Communications (obsolete)\n");
    break;
  case 0x0A:
  case 0x0B:
    fprintf(fp, " (Graphics arts pre-press)\n");
    break;
  case 0x0C:
    fprintf(fp, " Storage array controller (RAID)\n");
    break;
  case 0x0D:
    fprintf(fp, " Enclosure services\n");
    break;
  case 0x0E:
    fprintf(fp, " Simplified direct-access (magnetic disk)\n");
    break;
  case 0x0F:
    fprintf(fp, " Optical card reader/writer\n");
    break;
  case 0x10:
    fprintf(fp, " (Reserved by Bridging Expanders)\n");
    break;
  case 0x11:
    fprintf(fp, " Object-based storage\n");
    break;
  case 0x12:
    fprintf(fp, " Automation/Drive Interface\n");
    break;
  case 0x1E:
    fprintf(fp, " Well known logical unit\n");
    break;
  case 0x1F:
    fprintf(fp, " Unknown or no device type\n");
    break;
  default:
    fprintf(fp, " Reserved\n");
    break;
    
  }

  fprintf(fp, "RMB:");
  if(data[1]>>7) {
    fprintf(fp, " medium is removable\n");
  } else {
    fprintf(fp, " medium is not removable\n");
  }
  
  fprintf(fp, "Version: ");
  switch(data[2]) {
  case 0x00:
    fprintf(fp, "The device does not claim conformance to any standard\n");
    break;
  case 0x01:
    fprintf(fp, "(%02x)Obsolete (SCSI = 1)\n", data[2]);
    break;
  case 0x02:
    fprintf(fp, "(%02x) Obsolete\n", data[2]);
    break;
  case 0x03:
    fprintf(fp, "Complies to ANSI X3.301:1997 (SPC)\n");
    break;
  case 0x04:
    fprintf(fp, "Complies to ANSI NCITS.351:2001 (SPC-2)\n");
    break;
  case 0x05:
    fprintf(fp, "Complies to xxxxxxx (SPC-3)\n");
    break;
  case 0x08:
  case 0x09:
  case 0x0A:
  case 0x0B:
  case 0x0C:
    fprintf(fp, "(%02x) Obsolete (ECMA = 1)\n", data[2]);
    break;
  case 0x40:
  case 0x41:
  case 0x42:
  case 0x43:
  case 0x44:
    fprintf(fp, "(%02x) Obsolete (ISO = 1)\n", data[2]);
    break;
  case 0x48:
  case 0x49:
  case 0x4A:
  case 0x4B:
  case 0x4C:
    fprintf(fp, "(%02x) Obsolete (ISO = 1, ECMA = 1)\n", data[2]);
    break;
  case 0x80:
  case 0x81:
  case 0x82:
  case 0x83:
  case 0x84:
    fprintf(fp, "(%02x) Obsolete (ISO = 2)\n", data[2]);
    break;
  case 0x88:
  case 0x89:
  case 0x8A:
  case 0x8B:
  case 0x8C:
    fprintf(fp, "(%02x) Obsolete (ECMA = 1)\n", data[2]);
    break;
  default:
    fprintf(fp, "(%02x) Reserved\n", data[2]);
    break;
  }

  fprintf(fp, "byte 3 *not implemented*\n");
  fprintf(fp, "byte 4 *not implemented*\n");
  fprintf(fp, "byte 5 *not implemented*\n");
  fprintf(fp, "byte 6 *not implemented*\n");
  fprintf(fp, "byte 7 *not implemented*\n");

  fprintf(fp, "Vendor Identification: '%.8s'\n", &data[8]);
  fprintf(fp, "Product Identification: '%.16s'\n", &data[16]);
  fprintf(fp, "Product Revision Level: '%.4s'\n", &data[32]);
  return;
}


void print_PhysicalFormatInformation(FILE *fp,
				     PhysicalFormatInformation_t *info)
{
  int n;
  n = 4;

  fprintf(fp, "Physical Format Information\n");
  
  fprintf(fp, "Book Type:");
  switch(info->b[n] >> 4) {
  case 0:
    fprintf(fp, " DVD-ROM");
    break;
  case 1:
    fprintf(fp, " DVD-RAM");
    break;
  case 2:
    fprintf(fp, " DVD-R");
    break;
  case 3:
    fprintf(fp, " DVD-RW");
    break;
  case 9:
    fprintf(fp, " DVD+RW");
    break;
  default:
    fprintf(fp, " %u (Reserved)", info->b[n] >> 4);
    break;
  }
  
  fprintf(fp, ", Part Version: %u", info->b[n] & 0x0f);
  n++;
  
  
  fprintf(fp, "\nDisc Size:");
  switch(info->b[n] >> 4) {
  case 0:
    fprintf(fp, " 120mm");
    break;
  case 1:
    fprintf(fp, " 80mm");
    break;
  default:
    fprintf(fp, " %u (Reserved)", info->b[n] >> 4);
    break;
  }

  fprintf(fp, ", Maximum Rate:");
  switch(info->b[n] & 0x0f) {
  case 0:
    fprintf(fp, " 2.52 Mbps");
    break;
  case 1:
    fprintf(fp, " 5.04 Mbps");
    break;
  case 2:
    fprintf(fp, " 10.08 Mbps");
    break;
  case 15:
    fprintf(fp, " Not Specified");
    break;
  default:
    fprintf(fp, " %u (Reserved)", info->b[n] & 0x0f);
    break;
  }
  n++;
  


  
  fprintf(fp, "\nNumber of Layers:");
  switch((info->b[n] >> 5) & 0x3) {
  case 0:
    fprintf(fp, " 1");
    break;
  case 1:
    fprintf(fp, " 2");
    break;
  default:
    fprintf(fp, " %u (Reserved)", (info->b[n] >> 5) & 0x3);
    break;
  }


  fprintf(fp, ", Track Path:");
  switch((info->b[n] >> 4) & 0x1) {
  case 0:
    fprintf(fp, " Parallel");
    break;
  case 1:
    fprintf(fp, " Opposite");
    break;
  default:
    break;
  }


  fprintf(fp, ", Layer Type:");
  switch(info->b[n] & 0x0f) {
  case 1:
    fprintf(fp, " Read-only");
    break;
  case 2:
    fprintf(fp, " Recordable");
    break;
  case 4:
    fprintf(fp, " Rewritable");
    break;
  default:
    fprintf(fp, " %u (Reserved)", info->b[n] & 0x0f);
    break;
  }
  n++;

  
  fprintf(fp, "\nLinear Density:");
  switch(info->b[n] >> 4) {
  case 0:
    fprintf(fp, " 0.267 um/bit");
    break;
  case 1:
    fprintf(fp, " 0.293 um/bit");
    break;
  case 2:
    fprintf(fp, " 0.409 - 0.435 um/bit");
    break;
  case 4:
    fprintf(fp, " 0.280 - 0.291 um/bit");
    break;
  case 8:
    fprintf(fp, " 0.353 um/bit");
    break;
  default:
    fprintf(fp, " %u (Reserved)", info->b[n] >> 4);
    break;
  }
  
  
  fprintf(fp, ", Track Density:");
  switch(info->b[n] & 0x0f) {
  case 0:
    fprintf(fp, " 0.74 um/track");
    break;
  case 1:
    fprintf(fp, " 0.80 um/track");
    break;
  case 2:
    fprintf(fp, " 0.615 um/track");
    break;
  default:
    fprintf(fp, " %u (Reserved)", info->b[n] & 0x0f);
    break;
  }
  
  n++;
  
  // 0-byte
  n++;
  
  fprintf(fp, "\nStarting PSN of Data Area: 0x%08x",
	  (info->b[n] << 16) | (info->b[n+1] << 8) | (info->b[n+2]));
  n+=3;
  
  // 0-byte
  
  n++;
  
  fprintf(fp, "\nEnd PSN of Data Area: 0x%08x",
	  (info->b[n] << 16) | (info->b[n+1] << 8) | (info->b[n+2]));
  n+=3;
  
  // 0-byte
  n++;
  
  fprintf(fp, "\nEnd Sector Number in Layer 0: 0x%08x",
	  (info->b[n] << 16) | (info->b[n+1] << 8) | (info->b[n+2]));
  n+=3;
  
  fprintf(fp, "\nBCA Flag: %u", info->b[n] >> 7);
  n++;
  
  fprintf(fp, "\nMedia Specific: (*Not Implemented*)\n");

}


void print_CopyrightInformation(FILE *fp, CopyrightInformation_t *info)
{

  fprintf(fp, "Copyright Information\n");
  fprintf(fp, "Copyright Protection System Type: ");
  switch(info->b[4]) {
  case 0:
    fprintf(fp, "None\n");
    break;
  case 1:
    fprintf(fp, "CSS/CPPM\n");
    break;
  case 2:
    fprintf(fp, "CPRM\n");
    break;
  default:
    fprintf(fp, "%02x (Reserved)\n", info->b[4]);
    break;
  }
  fprintf(fp, "Region Management Information: %02x\n",
	  info->b[5]);
  
  return;
}


void print_DiscManufacturingInformation(FILE *fp, 
					DiscManufacturingInformation_t *info)
{
  int n, m;
  int data_len;

  data_len = info->b[0]<<8 | info->b[1];
  fprintf(fp, "Disc Manufacturing Information\n");
  
  for(n = 4, m = 0; n < data_len+2; n++, m++) {
    if(m%16 == 0) {
      fprintf(fp, "\n");
    }
    if(isprint(info->b[n])) {
      fprintf(fp, "'%c',", info->b[n]);
    } else {
      fprintf(fp, " %02x,", info->b[n]);
    }
  }
  fprintf(fp, "\n");
  return;
}

  



void test(int fd)
{
  uint8_t cdb[12];
  uint8_t sense[18];
  uint8_t data[1000];
  userscsi_cmd_t cmd;
  int n;

  memset(&cmd, 0, sizeof(cmd));
  memset(cdb, 0, sizeof(cdb));
  memset(sense, 0, sizeof(sense));
  memset(data, 0, sizeof(data));
  cdb[0] = 0xAD;
  cdb[6] = 0; // layer
  cdb[7] = 0;
  cdb[8] = sizeof(data) >> 8;
  cdb[9] = sizeof(data) & 0xff;

  cmd.flags = USERSCSI_CMD_READ  | USERSCSI_CMD_RQENABLE;
  cmd.cdb = cdb;
  cmd.cdb_len = 12;
  cmd.data = data;
  cmd.data_len = sizeof(data);
  cmd.rs = sense;
  cmd.rs_len = sizeof(sense);
  cmd.timeout = 5;

  
  if(userscsi_cmd(fd, &cmd) == -1) {
    fprintf(stdout, "ERR: error\n");
  }
  fprintf(stdout, "ERR: status: %02x\n", cmd.status);
  fprintf(stdout, "ERR: request sense: %02x\n", cmd.rs[0]);

  for(n = 0; n < 18; n++) {
    fprintf(stdout, " %02x", cmd.rs[n]);
  }
  fprintf(stdout, "\n");

  fprintf(stdout, "%02x %02x %02x %02x\n",
	  data[0], data[1], data[2], data[3]);

}


void userscsi_psense(uint8_t *sense, int sense_len)
{
  uint8_t response_code;
  int additional_sense_len;

  if(sense_len < 8) {
    fprintf(stdout, "ERR: sense_len %d < 8\n", sense_len);
    return;
  }

  response_code = sense[0] & 0x7f;
  if(response_code == 0x70 || response_code == 0x71) {
#if 0
    fprintf(stdout, " response code: 0x%02x, ", response_code);
    fprintf(stdout, "ILI: %01x, ", (sense[2]>>5) & 0x01);
    fprintf(stdout, "Valid: %01x, ", sense[0] >> 7);
    fprintf(stdout, "Information: 0x%02x%02x%02x%02x\n",
	    sense[3], sense[4], sense[5], sense[6]); 
#endif
    additional_sense_len = sense[7];
    
    if(8+additional_sense_len < 14) {
      fprintf(stdout, "ERR: sense key: (0x%02x) %s\n",
	      sense[2] & 0x0f, userscsi_sense_key_str(sense[2]&0x0f));
      return;
    } else {
      
      fprintf(stdout, "ERR: sense: (%02x/%02x/%02x), %s, %s\n",
	      sense[2] & 0x0f, sense[12], sense[13],
	      userscsi_sense_key_str(sense[2]&0x0f),
	      userscsi_asc_ascq_str(sense[12], sense[13]));

    }

    if(8+additional_sense_len < 18)
      return;

#if 0
    fprintf(stdout, "ERR: Field replaceable unit code: 0x%02x\n",
	    sense[14]);
    
    fprintf(stdout, "ERR: SKSV: %01x, ", sense[15] >> 7);

    fprintf(stdout, "Sense-key specific: 0x%02x%02x%02x\n",
	    sense[15] & 0x7f, sense[16], sense[17]);
#endif

  } else {
    fprintf(stdout, "ERR: response code: %#02x\n", response_code);
  }


}

void userscsi_perror(int ret, uint8_t status,
		     uint8_t *sense, int sense_len,
		     char *str)
{
  
  if((ret == -1) && (errno != EIO)) {
    fprintf(stdout, "ERR: %s: %s\n", str, strerror(errno));
  } else {
    fprintf(stdout, "ERR: %s:\n", str);
  }
  
  fprintf(stdout, "ERR: Status: (0x%02x) %s\n", status,
	  userscsi_status_str(status));
  
  userscsi_psense(sense, sense_len);
  
}  




int print_char(uint8_t *data, int data_len)
{
  int n;

  for(n = 0; n < data_len; n++) {
    fprintf(stdout, "%c,", data[n]);
  }
  fprintf(stdout, "\n");

  return 0;
}

int print_hex(uint8_t *data, int data_len)
{
  int n;

  for(n = 0; n < data_len; n++) {
    fprintf(stdout, "%02x,", data[n]);
  }
  fprintf(stdout, "\n");

  return 0;
}

int main(int argc, char **argv)
{
  int fd;
  int ret;
  uint8_t data[4096];
  char tmp_str[256];
  
  if(argc != 2) {
    fprintf(stderr, "usage: dvd_test <device>\n");
    exit(1);
  }

  fprintf(stdout, "\ndvd_test output: (%s)\n", argv[1]);
  fprintf(stdout, "================\n\n");

  system("uname -mrsv");
  if(strlen(argv[1]) < (256-40)) {
    strcpy(tmp_str, "ls -l ");
    strcat(tmp_str, argv[1]);
    system(tmp_str);
    strcpy(tmp_str, "ls -lL ");
    strcat(tmp_str, argv[1]);
    system(tmp_str);
#if defined(__linux__)
    strcpy(tmp_str, "/sbin/blockdev --getsize --getss ");
    strcat(tmp_str, argv[1]);
    system(tmp_str);
#endif
  }

  fd = open(argv[1], O_RDONLY);
  
  if(fd == -1) {
    perror(argv[1]);
    exit(1);
  }

  init_userscsi();

  
  memset(data, 0, sizeof(data));

  fprintf(stdout, "--- Inquiry ---\n");

  ret = Inquiry(fd, data, 56);
  if(ret < 0) {
    userscsi_perror(ret, status, request_sense_data, request_sense_data_len,
		    "Inquiry");
  } else {
    print_InquiryData(stdout, data, 56);
  }


  fprintf(stdout, "--- GetConfiguration 88 ---\n");
  
  ret = GetConfiguration(fd, 0, 0, data, 88);      
  if(ret < 0) {
    userscsi_perror(ret, status, request_sense_data, request_sense_data_len,
		    "GetConfiguration");
  } else {
    print_ConfigurationData(stdout, data, 88);
  }
  
  {
    int data_len;
    fprintf(stdout, "--- GetConfiguration 4 ---\n");
    ret = GetConfiguration(fd, 0, 0, data, 4);
    data_len = data[0]<<24 | data[1]<<16 | data[2]<<8 | data[3];
    if(ret < 0) {
      userscsi_perror(ret, status, request_sense_data, request_sense_data_len,
		      "GetConfiguration");
    } else {
      fprintf(stdout, "--- GetConfiguration %d+4 ---\n", data_len);

      ret = GetConfiguration(fd, 0, 0, data, data_len+4);      
      if(ret < 0) {
	userscsi_perror(ret,status, request_sense_data, request_sense_data_len,
			"GetConfiguration");
      } else {
	print_ConfigurationData(stdout, data, data_len+4);
      }
    }
    
  }

  fprintf(stdout, "--- GetConfiguration 256 ---\n");
  
  ret = GetConfiguration(fd, 0, 0, data, 256);      
  if(ret < 0) {
    userscsi_perror(ret, status, request_sense_data, request_sense_data_len,
		    "GetConfiguration");
  } else {
    print_ConfigurationData(stdout, data, 256);
  }


#if 1

  fprintf(stdout, "--- StructureList ---\n");

  ret = ReadDVDStructure_StructureList(fd, 0, data, sizeof(data));
  if(ret < 0) {
    userscsi_perror(ret, status, request_sense_data, request_sense_data_len,
		    "ReadDVDStructure_StructureList");
    ret = Inquiry_CmdDt(fd, data, 6, 0x12);
    if(ret < 0) {
      userscsi_perror(ret, status, request_sense_data, request_sense_data_len,
		      "Inquiry_CmdDt");
    } else {
      print_CommandSupportData(stdout, data, 6);
    }
  } else {
    StructureListEntry_t *sl;
    int data_len;
    int nr_sl;
    int n;
    data_len = data[0]<<8 | data[1];
    
    fprintf(stdout, "Structure List data len: %02x\n", data_len);
    
    if(data_len > 5) {
      nr_sl = (data_len-2)/sizeof(*sl);
    } else {
      nr_sl = 0;
    }
    
    sl = (StructureListEntry_t *)&data[4];
    
    for(n = 0; n < nr_sl; n++) {
      fprintf(stdout, "Format Code: %02x\n", sl[n].b[0]);
      fprintf(stdout, "SDS: %01x, RDS: %01x\n",
	      sl[n].b[1]>>7, sl[n].b[1]>>6 & 0x1);
      fprintf(stdout, "Structure Length: %u\n", sl[n].b[2]<<8 | sl[n].b[3]);
    }
    
  }
#endif

  fprintf(stdout, "--- Physical ---\n");

  ret = ReadDVDStructure_Physical(fd, 0, 
				  (PhysicalFormatInformation_t *)data);
  if(ret < 0) {
    userscsi_perror(ret, status, request_sense_data, request_sense_data_len,
		    "ReadDVDStructure_Physical");
  } else { 

    print_PhysicalFormatInformation(stdout,
				    (PhysicalFormatInformation_t *)data);
  }

  fprintf(stdout, "--- Copyright ---\n");
  
  ret = ReadDVDStructure_Copyright(fd, 0, (CopyrightInformation_t *)data);
  if(ret < 0) {
    userscsi_perror(ret, status, request_sense_data, request_sense_data_len,
		    "ReadDVDStructure_Copyright");
  } else {

    print_CopyrightInformation(stdout, (CopyrightInformation_t *)data);
  
  }

  fprintf(stdout, "--- RPC State ---\n");

  ret = ReportKey_RPCState(fd, data, 8);
  if(ret < 0) {
    userscsi_perror(ret, status, request_sense_data, request_sense_data_len,
		    "ReportKey - RPC State");
  } else {
    print_RPCState(stdout, data, 8);
  }
  
#if 0
  fprintf(stdout, "--- Manufacturing ---\n");

  ret = ReadDVDStructure_Manufacturing(fd, 0,
				       (DiscManufacturingInformation_t *)data);
  if(ret < 0) {
    userscsi_perror(ret, status, request_sense_data, request_sense_data_len,
		    "ReadDVDStructure_Manufacturing");
  } else {
    
    print_DiscManufacturingInformation(stdout, 
				       (DiscManufacturingInformation_t *)data);
  }
#endif

  fprintf(stdout, "--- Capacity ---\n");

  ret = ReadCapacity(fd, data, 8);
  if(ret < 0) {
    userscsi_perror(ret, status, request_sense_data, request_sense_data_len,
		    "ReadCapacity");
  } else {
    print_CapacityData(stdout, (CapacityData_t *)data);
  }


  fprintf(stdout, "--- Read ---\n");

  {
    uint32_t lba[] = {
      0, 16, 256, 512, 1024, 2048, 10000
    };
    int total_lba = sizeof(lba)/sizeof(uint32_t);
    int n;
    
    for(n = 0; n < total_lba; n++) {
      
      ret = Read_10(fd, lba[n], 1, 0, data, 2048);
      if(ret < 0) {
	userscsi_perror(ret,status, request_sense_data, request_sense_data_len,
			"Read_10");
	fprintf(stdout, "ERR: lba: %u\n", lba[n]);
      } else {
	fprintf(stdout, "success reading lba %u with Read_10\n", lba[n]);
	if(lba[n] == 16) {
	  print_char(data, 128);
	}
	
	if(lba[n] == 256) {
	  print_hex(data, 4);
	}
      }



      ret = Read_12(fd, lba[n], 1, 0, data, 2048);
      if(ret < 0) {
	userscsi_perror(ret,status, request_sense_data, request_sense_data_len,
			"Read_12");
	fprintf(stdout, "ERR: lba: %u\n", lba[n]);
      } else {
	fprintf(stdout, "success reading lba %u with Read_12\n", lba[n]);
	if(lba[n] == 16) {
	  print_char(data, 1024);
	}
	
	if(lba[n] == 256) {
	  print_hex(data, 4);
	}
      }
      
    }

  }



  fprintf(stdout, "--- read ---\n");

  {
    int n;
    
    for(n = 0; n < 257; n++) {
      
      ret = read(fd, data, 2048);
      if(ret == -1) {
	fprintf(stdout, "ERR: read: %s\n", strerror(errno));
	fprintf(stdout, "ERR: lba: %u\n", n);
      } else if (ret != 2048) {
	fprintf(stdout, "ERR: only got %d bytes, lba: %u\n", ret, n);
      } else if(n == 16) {
	print_char(data, 4);
      } else if(n == 256) {
	print_hex(data, 4);
      }
    }
    
  }
  
  
  fprintf(stdout, "--- seek read ---\n");
  
  {
    uint32_t lba[] = {
      0, 16, 256, 512, 1024, 2048, 10000
    };
    int total_lba = sizeof(lba)/sizeof(uint32_t);
    int n;
    
    for(n = 0; n < total_lba; n++) {
      
      if((ret = lseek(fd, lba[n]*2048, SEEK_SET)) == -1) {
	fprintf(stdout, "ERR: lseek: %s\n", strerror(errno));
	fprintf(stdout, "ERR: lba: %u\n", lba[n]);
      } else {
	fprintf(stdout, "success seeking to lba %u\n", lba[n]);
	
	ret = read(fd, data, 2048);
	if(ret == -1) {
	  fprintf(stdout, "ERR: read: %s\n", strerror(errno));
	  fprintf(stdout, "ERR: lba: %u\n", lba[n]);
	} else if (ret != 2048) {
	  fprintf(stdout, "ERR: only got %d bytes, lba: %u\n", ret, lba[n]);
	} else {
	  fprintf(stdout, "success reading lba %u\n", lba[n]);
	}
      }
    }
    
  }
  return 0;
}

