/*
 * Copyright (C) 2002 Bjrn Englund
 */

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include "userscsi.h"

#if defined(__linux__)
static int clk_tck = 100; //some default value in case sysconf doesn't work;
#endif

int init_userscsi(void)
{
#if defined(__linux__)
  int tck;
  tck = sysconf(_SC_CLK_TCK);
  if(tck > 0) {
    clk_tck = tck;
    fprintf(stdout, "clk_tck: %d\n", tck);
  } else {
    perror("sysconf(_SC_CLK_TCK)");
  }
#endif
  return 0;
}



int userscsi_cmd(int fd, userscsi_cmd_t *cmd)
{
#if defined(__sun__)
  struct uscsi_cmd solaris_cmd;  
  int ret;
  
  memset(&solaris_cmd, 0, sizeof(solaris_cmd));
  
  solaris_cmd.uscsi_flags = cmd->flags;
  solaris_cmd.uscsi_cdb = cmd->cdb;
  solaris_cmd.uscsi_cdblen = cmd->cdb_len;
  solaris_cmd.uscsi_bufaddr = cmd->data;
  solaris_cmd.uscsi_buflen = cmd->data_len;
  solaris_cmd.uscsi_rqbuf = cmd->rs;
  solaris_cmd.uscsi_rqlen = cmd->rs_len;
  solaris_cmd.uscsi_timeout = cmd->timeout;


  ret = ioctl(fd, USCSICMD, &solaris_cmd);

  cmd->status = solaris_cmd.uscsi_status;
  cmd->data_resid = solaris_cmd.uscsi_resid;
  cmd->rs_status = solaris_cmd.uscsi_rqstatus;
  cmd->rs_resid = solaris_cmd.uscsi_rqresid;

  return ret;
  
#elif defined(__linux__)
  struct cdrom_generic_command linux_cmd;
  struct request_sense linux_sense;
  int quiet;
  uint8_t direction;
  int request_sense_len = 0;
  int ret;
  
  if(cmd->cdb_len > CDROM_PACKET_SIZE) {
    // Linux supports max 12 byte commands at this point
    // in fact it always sends 12 bytes 
    // hopefully 0-padded as in the mmc spec
    fprintf(stdout, "ERR: userscsi_cmd: cdb_len %d > CDROM_PACKET_SIZE: %d\n",
	    cmd->cdb_len, CDROM_PACKET_SIZE);
    return -1;
  }
  
  
  memset(&linux_cmd, 0, sizeof(linux_cmd));
  memset(&linux_sense, 0, sizeof(linux_sense));
  
  if(cmd->flags & USERSCSI_CMD_READ) {
    direction = CGC_DATA_READ;
  } else if(cmd->data_len > 0) {
    direction = CGC_DATA_WRITE;
  } else {
    direction = CGC_DATA_NONE;
  }

  if(cmd->flags & USERSCSI_CMD_SILENT) {
    quiet = 1;
  } else {
    quiet = 0;
  }

  memcpy(linux_cmd.cmd, cmd->cdb, cmd->cdb_len);
  linux_cmd.buffer = cmd->data;
  linux_cmd.buflen = cmd->data_len;
  if(cmd->rs_len < sizeof(linux_sense)) {
    // Linux always returns a 64 byte long struct with sense data
    // I wonder how many of those bytes are really filled in?
    linux_cmd.sense = &linux_sense;
  } else {
    linux_cmd.sense = (struct request_sense *)cmd->rs;
  }
  
  linux_cmd.data_direction = direction;
  linux_cmd.quiet = quiet;
  linux_cmd.timeout = cmd->timeout*clk_tck;

  cmd->status = 0;
  
  ret = ioctl(fd, CDROM_SEND_PACKET, &linux_cmd);

  fprintf(stdout, "ERR: ret: %d, status: %d\n", ret, linux_cmd.stat);
  if(ret == -1) {
    if(errno == EIO) {
      cmd->status = 2;
    } else {
      fprintf(stdout, "ERR: errno != EIO\n");
      cmd->status = 2;
    }
  }
  
  if(cmd->status != 0) {
    request_sense_len = linux_cmd.sense->add_sense_len + 8;
    if(cmd->rs_len < sizeof(linux_sense)) {
      if(request_sense_len < cmd->rs_len) {
	memcpy(cmd->rs, linux_cmd.sense, request_sense_len);
      } else {
	memcpy(cmd->rs, linux_cmd.sense, cmd->rs_len);
      }
    }
  }
  
  return ret;
#elif defined(__bsdi__)
  struct scsi_user_cdb suc;
  
  bzero(&suc, sizeof (suc));
  if(cmd->flags & USERSCSI_CMD_READ) {
    suc.suc_flags = SUC_READ;
  } else {
    suc.suc_flags = SUC_WRITE;
  }
  suc.suc_cdblen = cmd->cdb_len;
  bcopy(cmd->cdb, suc.suc_cdb, cmd->cdb_len);
  suc.suc_data = cmd->data;
  suc.suc_datalen = cmd->data_len;
  suc.suc_timeout = cmd->timeout;
  
  if(ioctl(fd, SCSIRAWCDB, &suc) == -1) {
    return(-1);
  }
  cmd->status = suc.suc_sus.sus_status;
  if(cmd->status) {
    bcopy(&suc.suc_sus.sus_sense, cmd->rs, cmd->rs_len);
  }
  cmd->data_resid = cmd->data_len - suc.suc_datalen;
  return(0);
  
#endif
} 
