/*
 * Copyright (C) 2002 Bjrn Englund
 */

#include <stdlib.h>
#include <inttypes.h>

const char reserved_str[] = "(Reserved)";
const char null_str[] = "(NULL)";

#define STATUS_ENTRIES 0x41
const char *status_str[STATUS_ENTRIES] = {
  /*0x00*/ "GOOD",
  /*0x01*/ reserved_str,
  /*0x02*/ "CHECK CONDITION",
  /*0x03*/ reserved_str,
  /*0x04*/ "CONDITION MET",
  /*0x05*/ reserved_str,
  /*0x06*/ reserved_str,
  /*0x07*/ reserved_str,
  /*0x08*/ "BUSY",
  /*0x09*/ reserved_str,
  /*0x0A*/ reserved_str,
  /*0x0B*/ reserved_str,
  /*0x0C*/ reserved_str,
  /*0x0D*/ reserved_str,
  /*0x0E*/ reserved_str,
  /*0x0F*/ reserved_str,
  /*0x10*/ "INTERMEDIATE",
  /*0x11*/ reserved_str,
  /*0x12*/ reserved_str,
  /*0x13*/ reserved_str,
  /*0x14*/ "INTERMEDIATE CONDITION MET",
  /*0x15*/ reserved_str,
  /*0x16*/ reserved_str,
  /*0x17*/ reserved_str,
  /*0x18*/ "RESERVATION CONFLICT",
  /*0x19*/ reserved_str,
  /*0x1A*/ reserved_str,
  /*0x1B*/ reserved_str,
  /*0x1C*/ reserved_str,
  /*0x1D*/ reserved_str,
  /*0x1E*/ reserved_str,
  /*0x1F*/ reserved_str,
  /*0x20*/ reserved_str,
  /*0x21*/ reserved_str,
  /*0x22*/ "Obsolete",
  /*0x23*/ reserved_str,
  /*0x24*/ reserved_str,
  /*0x25*/ reserved_str,
  /*0x26*/ reserved_str,
  /*0x27*/ reserved_str,
  /*0x28*/ "TASK SET FULL",
  /*0x29*/ reserved_str,
  /*0x2A*/ reserved_str,
  /*0x2B*/ reserved_str,
  /*0x2C*/ reserved_str,
  /*0x2D*/ reserved_str,
  /*0x2E*/ reserved_str,
  /*0x2F*/ reserved_str,
  /*0x30*/ "ACA ACTIVE",
  /*0x31*/ reserved_str,
  /*0x32*/ reserved_str,
  /*0x33*/ reserved_str,
  /*0x34*/ reserved_str,
  /*0x35*/ reserved_str,
  /*0x36*/ reserved_str,
  /*0x37*/ reserved_str,
  /*0x38*/ reserved_str,
  /*0x39*/ reserved_str,
  /*0x3A*/ reserved_str,
  /*0x3B*/ reserved_str,
  /*0x3C*/ reserved_str,
  /*0x3D*/ reserved_str,
  /*0x3E*/ reserved_str,
  /*0x3F*/ reserved_str,
  /*0x40*/ "TASK ABORTED"
};


#define SENSE_KEY_ENTRIES 0x10
const char *sense_key_str[SENSE_KEY_ENTRIES] = {
  "NO SENSE",
  "RECOVERED ERROR",
  "NOT READY",
  "MEDIUM ERROR",
  "HARDWARE ERROR",
  "ILLEGAL REQUEST",
  "UNIT ATTENTION",
  "DATA PROTECT",
  "BLANK CHECK",
  "VENDOR SPECIFIC",
  "COPY ABORTED",
  "ABORTED COMMAND",
  "(Obsolete)",
  "VOLUME OVERFLOW",
  "MISCOMPARE",
  "(Reserved)"
};

typedef struct {
  uint8_t nr_entries;
  const char **str;
} asc_entry_t;

/*** SPC-3 ***/
#define ASC_00_ENTRIES 0x1d
const char *asc_00_str[ASC_00_ENTRIES] = {
  /* 0x00 */ "NO ADDITIONAL SENSE INFORMATION",
  /* 0x01 */ "FILEMARK DETECTED",
  /* 0x02 */ "END-OF-PARTITION/MEDIUM DETECTED",
  /* 0x03 */ "SETMARK DETECTED",
  /* 0x04 */ "BEGINNING-OF-PARTITION/MEDIUM DETECTED",
  /* 0x05 */ "END-OF-DATA DETECTED",
  /* 0x06 */ "I/O PROCESS TERMINATED",
  /* 0x07 */ reserved_str,
  /* 0x08 */ reserved_str,
  /* 0x09 */ reserved_str,
  /* 0x0A */ reserved_str,
  /* 0x0B */ reserved_str,
  /* 0x0C */ reserved_str,
  /* 0x0D */ reserved_str,
  /* 0x0E */ reserved_str,
  /* 0x0F */ reserved_str,
  /* 0x10 */ reserved_str,
  /* 0x11 */ "AUDIO PLAY OPERATION IN PROGRESS",
  /* 0x12 */ "AUDIO PLAY OPERATION PAUSED",
  /* 0x13 */ "AUDIO PLAY OPERATION SUCCESSFULLY COMPLETED",
  /* 0x14 */ "AUDIO PLAY OPERATION STOPPED DUE TO ERROR",
  /* 0x15 */ "NO CURRENT AUDIO STATUS TO RETURN",
  /* 0x16 */ "OPERATION IN PROGRESS",
  /* 0x17 */ "CLEANING REQUESTED",
  /* 0x18 */ "ERASE OPERATION IN PROGRESS",
  /* 0x19 */ "LOCATE OPERATION IN PROGRESS",
  /* 0x1A */ "REWIND OPERATION IN PROGRESS",
  /* 0x1B */ "SET CAPACITY OPERATION IN PROGRESS",
  /* 0x1C */ "VERIFY OPERATION IN PROGRESS"
};


/*** SPC-3 ***/

#define ASC_24_ENTRIES 0x04
const char *asc_24_str[ASC_24_ENTRIES] = {
    /* 0x00 */ "INVALID FIELD IN CDB",
    /* 0x01 */ "CDB DECRYPTION ERROR",
    /* 0x02 */ "Obsolete SPC-2 (INVALID CDB FIELD WHILE IN EXPLICIT BLOCK ADDRESS MODEL)",
    /* 0x03 */ "Obsolete SPC-2 (INVALID CDB FIELD WHILE IN IMPLICIT BLOCK ADDRESS MODEL)"
};


/*** SPC-3 ***/
#define ASC_30_ENTRIES 0x11
const char *asc_30_str[ASC_30_ENTRIES] = {
  /* 0x00 */ "INCOMPATIBLE MEDIUM INSTALLED",
  /* 0x01 */ "CANNOT READ MEDIUM - UNKNOWN FORMAT",
  /* 0x02 */ "CANNOT READ MEDIUM - INCOMPATIBLE FORMAT",
  /* 0x03 */ "CLEANING CARTRIDGE INSTALLED",
  /* 0x04 */ "CANNOT WRITE MEDIUM - UNKNOWN FORMAT",
  /* 0x05 */ "CANNOT WRITE MEDIUM - INCOMPATIBLE FORMAT",
  /* 0x06 */ "CANNOT FORMAT MEDIUM - INCOMPATIBLE MEDIUM",
  /* 0x07 */ "CLEANING FAILURE",
  /* 0x08 */ "CANNOT WRITE - APPLICATION CODE MISMATCH",
  /* 0x09 */ "CURRENT SESSION NOT FIXATED FOR APPEND",
  /* 0x0A */ reserved_str,
  /* 0x0B */ reserved_str,
  /* 0x0C */ reserved_str,
  /* 0x0D */ reserved_str,
  /* 0x0E */ reserved_str,
  /* 0x0F */ reserved_str,
  /* 0x10 */ "MEDIUM NOT FORMATTED"
};


/*** SPC-3 ***/
#define ASC_3A_ENTRIES 0x05
const char *asc_3A_str[ASC_3A_ENTRIES] = {
  /* 0x00 */ "MEDIUM NOT PRESENT",
  /* 0x01 */ "MEDIUM NOT PRESENT - TRAY CLOSED",
  /* 0x02 */ "MEDIUM NOT PRESENT - TRAY OPEN",
  /* 0x03 */ "MEDIUM NOT PRESENT - LOADABLE",
  /* 0x04 */ "MEDIUM NOT PRESENT - MEDIUM AUXILIARY MEMORY ACCESSIBLE"
};


#define ASC_ASCQ_ENTRIES 0x40
asc_entry_t asc_ascq_str[ASC_ASCQ_ENTRIES] = {
  /*0x00*/ { ASC_00_ENTRIES, asc_00_str},
  /*0x01*/ { 0, NULL},
  /*0x02*/ { 0, NULL},
  /*0x03*/ { 0, NULL},
  /*0x04*/ { 0, NULL},
  /*0x05*/ { 0, NULL},
  /*0x06*/ { 0, NULL},
  /*0x07*/ { 0, NULL},
  /*0x08*/ { 0, NULL},
  /*0x09*/ { 0, NULL},
  /*0x0A*/ { 0, NULL},
  /*0x0B*/ { 0, NULL},
  /*0x0C*/ { 0, NULL},
  /*0x0D*/ { 0, NULL},
  /*0x0E*/ { 0, NULL},
  /*0x0F*/ { 0, NULL},
  /*0x10*/ { 0, NULL},
  /*0x11*/ { 0, NULL},
  /*0x12*/ { 0, NULL},
  /*0x13*/ { 0, NULL},
  /*0x14*/ { 0, NULL},
  /*0x15*/ { 0, NULL},
  /*0x16*/ { 0, NULL},
  /*0x17*/ { 0, NULL},
  /*0x18*/ { 0, NULL},
  /*0x19*/ { 0, NULL},
  /*0x1A*/ { 0, NULL},
  /*0x1B*/ { 0, NULL},
  /*0x1C*/ { 0, NULL},
  /*0x1D*/ { 0, NULL},
  /*0x1E*/ { 0, NULL},
  /*0x1F*/ { 0, NULL},
  /*0x20*/ { 0, NULL},
  /*0x21*/ { 0, NULL},
  /*0x22*/ { 0, NULL},
  /*0x23*/ { 0, NULL},
  /*0x24*/ { ASC_24_ENTRIES, asc_24_str},
  /*0x25*/ { 0, NULL},
  /*0x26*/ { 0, NULL},
  /*0x27*/ { 0, NULL},
  /*0x28*/ { 0, NULL},
  /*0x29*/ { 0, NULL},
  /*0x2A*/ { 0, NULL},
  /*0x2B*/ { 0, NULL},
  /*0x2C*/ { 0, NULL},
  /*0x2D*/ { 0, NULL},
  /*0x2E*/ { 0, NULL},
  /*0x2F*/ { 0, NULL},
  /*0x30*/ { ASC_30_ENTRIES, asc_30_str},
  /*0x31*/ { 0, NULL},
  /*0x32*/ { 0, NULL},
  /*0x33*/ { 0, NULL},
  /*0x34*/ { 0, NULL},
  /*0x35*/ { 0, NULL},
  /*0x36*/ { 0, NULL},
  /*0x37*/ { 0, NULL},
  /*0x38*/ { 0, NULL},
  /*0x39*/ { 0, NULL},
  /*0x3A*/ { ASC_3A_ENTRIES, asc_3A_str},
  /*0x3B*/ { 0, NULL},
  /*0x3C*/ { 0, NULL},
  /*0x3D*/ { 0, NULL},
  /*0x3E*/ { 0, NULL},
  /*0x3F*/ { 0, NULL}
};


const char *userscsi_status_str(uint8_t status)
{
  if(status < STATUS_ENTRIES) {
    return status_str[status];
  } else {
    return reserved_str;
  }

}

const char *userscsi_sense_key_str(uint8_t sense_key)
{
  if(sense_key < SENSE_KEY_ENTRIES) {
    return sense_key_str[sense_key];
  } else {
    return null_str;
  }
}

const char *userscsi_asc_ascq_str(uint8_t asc, uint8_t ascq)
{
  if((asc >= ASC_ASCQ_ENTRIES) ||
     (asc_ascq_str[asc].str == NULL) ||
     (ascq >= asc_ascq_str[asc].nr_entries)) {
    return reserved_str;
  } else {
    return asc_ascq_str[asc].str[ascq];
  }

}
